from typing import Any, Dict, List

from ._base_type import OMEType
from .model.reference import Reference
from .model.simple_types import LSID


def collect_references(value: Any) -> List[Reference]:
    """Return a list of all References contained in value.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is either a Reference object, or an uninteresting type
    that we don't need to inspect further.

    """
    references: List[Reference] = []
    if isinstance(value, Reference):
        references.append(value)
    elif isinstance(value, list):
        for v in value:
            references.extend(collect_references(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            references.extend(collect_references(getattr(value, f)))
    # Do nothing for uninteresting types
    return references


def collect_ids(value: Any) -> Dict[LSID, OMEType]:
    """Return a map of all model objects contained in value, keyed by id.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is neither a dataclass nor a list.

    """
    ids: Dict[LSID, OMEType] = {}
    if isinstance(value, list):
        for v in value:
            ids.update(collect_ids(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            if f == "id" and not isinstance(value, Reference):
                # We don't need to recurse on the id string, so just record it
                # and move on.
                ids[value.id] = value  # type: ignore
            else:
                ids.update(collect_ids(getattr(value, f)))
    # Do nothing for uninteresting types.
    return ids



from ome_types._base_type import OMEType

from .basic_annotation import BasicAnnotation


class NumericAnnotation(BasicAnnotation, OMEType):
    """An abstract Numeric Annotation from which some others are derived.

    Parameters
    ----------
    id : AnnotationID
    annotator : ExperimenterID, optional
        The Annotator is the person who attached this annotation. e.g. If
        UserA annotates something with TagB, owned by UserB, UserA is still
        the Annotator.
    namespace : str, optional
        We recommend the inclusion of a namespace for annotations you define.
        If it is absent then we assume the annotation is to use our (OME's)
        default interpretation for this type.
    """



from ._units import ureg

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

try:
    from . import model
    from .model import OME
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Could not import 'ome_types.model.OME'.\nIf you are in a dev environment, "
        "you may need to run 'python -m src.ome_autogen'" + str(e)
    ) from None
from .schema import to_dict, to_xml, validate  # isort:skip
from ._convenience import from_tiff, from_xml  # isort:skip

__all__ = [
    "to_dict",
    "validate",
    "from_xml",
    "to_xml",
    "from_tiff",
    "OME",
    "model",
    "ureg",
]



import os
from pathlib import Path
from typing import Union

from .model import OME


def from_xml(xml: Union[Path, str]) -> OME:  # type: ignore
    """Generate OME metadata object from XML string or path.

    Parameters
    ----------
    xml : Union[Path, str]
        Path to an XML file, or literal XML string.

    Returns
    -------
    ome: ome_types.model.ome.OME
        ome_types.OME metadata object
    """
    from .schema import to_dict

    xml = os.fspath(xml)
    d = to_dict(xml)
    for key in list(d.keys()):
        if key.startswith(("xml", "xsi")):
            d.pop(key)

    return OME(**d)  # type: ignore


def from_tiff(path: Union[Path, str]) -> OME:
    """Generate OME metadata object from OME-TIFF path.

    This will use the first ImageDescription tag found in the TIFF header.

    Parameters
    ----------
    path : Union[Path, str]
        Path to OME TIFF.

    Returns
    -------
    ome: ome_types.model.ome.OME
        ome_types.OME metadata object

    Raises
    ------
    ValueError
        If the TIFF file has no OME metadata.
    """
    return from_xml(_tiff2xml(path))


def _tiff2xml(path: Union[Path, str]) -> str:
    """Extract OME XML from OME-TIFF path.

    This will use the first ImageDescription tag found in the TIFF header.

    Parameters
    ----------
    path : Union[Path, str]
        Path to OME TIFF.

    Returns
    -------
    xml : str
        OME XML

    Raises
    ------
    ValueError
        If the TIFF file has no OME metadata.
    """
    from struct import unpack

    with Path(path).open(mode="rb") as fh:
        try:
            offsetsize, offsetformat, tagnosize, tagnoformat, tagsize, codeformat = {
                b"II*\0": (4, "<I", 2, "<H", 12, "<H"),
                b"MM\0*": (4, ">I", 2, ">H", 12, ">H"),
                b"II+\0": (8, "<Q", 8, "<Q", 20, "<H"),
                b"MM\0+": (8, ">Q", 8, ">Q", 20, ">H"),
            }[fh.read(4)]
        except KeyError:
            raise ValueError(f"{path!r} does not have a recognized TIFF header")

        fh.read(4 if offsetsize == 8 else 0)
        fh.seek(unpack(offsetformat, fh.read(offsetsize))[0])
        for _ in range(unpack(tagnoformat, fh.read(tagnosize))[0]):
            tagstruct = fh.read(tagsize)
            if unpack(codeformat, tagstruct[:2])[0] == 270:
                size = unpack(offsetformat, tagstruct[4 : 4 + offsetsize])[0]
                if size <= offsetsize:
                    desc = tagstruct[4 + offsetsize : 4 + offsetsize + size]
                    break
                fh.seek(unpack(offsetformat, tagstruct[-offsetsize:])[0])
                desc = fh.read(size)
                break
        else:
            raise ValueError(f"No OME metadata found in file: {path}")
    if desc[-1] == 0:
        desc = desc[:-1]
    return desc.decode("utf-8")



from pathlib import Path

from quilt3 import Package


def download_test_resources() -> None:
    root = Path(__file__).parent.parent.parent
    resources = (root / "aicsimageio" / "aicsimageio" / "tests" / "resources").resolve()

    # Get the specific hash for test resources
    with open(root / "aicsimageio" / "scripts" / "TEST_RESOURCES_HASH.txt", "r") as f:
        top_hash = f.readline().strip()

    # Download test resources
    resources.mkdir(exist_ok=True)
    package = Package.browse(
        "aicsimageio/test_resources",
        "s3://aics-modeling-packages-test-resources",
        top_hash=top_hash,
    )
    package["resources"].fetch(resources)


if __name__ == "__main__":
    download_test_resources()



from typing import Any, Dict, List, Optional, Union

from pydantic import Field, validator

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .ellipse import Ellipse
from .label import Label
from .line import Line
from .mask import Mask
from .point import Point
from .polygon import Polygon
from .polyline import Polyline
from .rectangle import Rectangle
from .shape import Shape
from .simple_types import ROIID

_shape_types: Dict[str, type] = {
    "point": Point,
    "line": Line,
    "rectangle": Rectangle,
    "ellipse": Ellipse,
    "polyline": Polyline,
    "polygon": Polygon,
    "mask": Mask,
    "label": Label,
}


class ROI(OMEType):
    """A four dimensional 'Region of Interest'.

    If they are not used, and the Image has more than one plane, the entire set of
    planes is assumed to be included in the ROI. Multiple ROIs may be specified.

    Parameters
    ----------
    id : ROIID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the ROI.
    name : str, optional
        The Name identifies the ROI to the user.
    union : List[Shape], optional
    """

    id: ROIID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    union: List[Shape] = Field(default_factory=list)

    @validator("union", pre=True, each_item=True)
    def validate_union(cls, value: Union[Shape, Dict[Any, Any]]) -> Shape:
        if isinstance(value, Shape):
            return value
        elif isinstance(value, dict):
            try:
                _type = value.pop("_type")
            except KeyError:
                raise ValueError("dict initialization requires _type") from None
            try:
                shape_cls = _shape_types[_type]
            except KeyError:
                raise ValueError(f"unknown Shape type '{_type}'") from None
            return shape_cls(**value)
        else:
            raise ValueError("invalid type for union values")



from distutils.command.build_py import build_py as _build_py
from runpy import run_path

from setuptools import setup


class build_py(_build_py):
    run_path("src/ome_autogen.py", run_name="__main__")


setup(
    cmdclass={"build_py": build_py},
    use_scm_version={"write_to": "src/ome_types/_version.py"},
)
